""" Methods for test cases involving checking command-line interfaces

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..data_model import TestCase
from ..warnings import TestCaseWarning
from biosimulators_utils.simulator.environ import ENVIRONMENT_VARIABLES
import re
import subprocess
import warnings

__all__ = [
    'CliDisplaysHelpInline',
    'CliDescribesSupportedEnvironmentVariablesInline',
    'CliDisplaysVersionInformationInline',
]


class CliDisplaysHelpInline(TestCase):
    """ Test that a command-line interface provides inline help. """

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        self.get_simulator_docker_image(specifications)
        image_url = specifications['image']['url']

        cli = [cli] if cli else ['docker', 'run', '--tty', '--rm', image_url]

        result = subprocess.run(cli, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        log = result.stdout.decode() if result.stdout else ''
        supported = (
            '-i' in log
            and '-o' in log
        )
        if not supported:
            warnings.warn(('Command-line interfaces should display basic help when no arguments are provided.\n\n'
                           'The command-line interface displayed the following when no argument was provided:\n\n  {}'
                           ).format(log.replace('\n', '\n  ')),
                          TestCaseWarning)

        result = subprocess.run(cli + ['-h'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        log = result.stdout.decode() if result.stdout else ''
        supported = (
            'arguments' in log
            and '-i' in log
            and '--archive' in log
            and '-o' in log
            and '--out-dir' in log
        )
        if not supported:
            warnings.warn(('Command-line interface should support the `-h` option for displaying help inline.\n\n'
                           'The command-line interface displayed the following when executed with `-h`:\n\n  {}'
                           ).format(log.replace('\n', '\n  ')),
                          TestCaseWarning)

        result = subprocess.run(cli + ['--help'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        log = result.stdout.decode() if result.stdout else ''
        supported = (
            'arguments' in log
            and '-i' in log
            and '--archive' in log
            and '-o' in log
            and '--out-dir' in log
        )
        if not supported:
            warnings.warn(('Command-line interface should support the `--help` option for displaying help inline.\n\n'
                           'The command-line interface displayed the following when executed with `--help`:\n\n  {}'
                           ).format(log.replace('\n', '\n  ')),
                          TestCaseWarning)


class CliDescribesSupportedEnvironmentVariablesInline(TestCase):
    """ Test that the inline help for a command-line interface describes the environment variables that the
    simulator supports.
    """

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        self.get_simulator_docker_image(specifications)
        image_url = specifications['image']['url']

        cli = [cli] if cli else ['docker', 'run', '--tty', '--rm', image_url]
        result = subprocess.run(cli + [' -h'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        log = result.stdout.decode() if result.stdout else ''

        potentially_missing_env_vars = []
        for var in ENVIRONMENT_VARIABLES.values():
            if var.name not in log:
                potentially_missing_env_vars.append(var.name)

        if potentially_missing_env_vars:
            msg = ('The inline help for a command-line interface for a simulation tool should describe the '
                   'environment variables that the simulation tool supports.\n\n'
                   'The command-line interface does not describe the following standard environment '
                   'variables recognized by BioSimulators:\n  - {}\n\n'
                   'If the simulation tool implements these variables, they should be described in the inline help for '
                   'its command-line interface.\n\n'
                   'Note, support for these environment variables is optional. Simulation tools are not required to support '
                   'these variables.'
                   ).format('\n  - '.join("'" + var + "'" for var in sorted(potentially_missing_env_vars)))
            warnings.warn(msg, TestCaseWarning)


class CliDisplaysVersionInformationInline(TestCase):
    """ Test that a command-line interface provides version information inline. """

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        self.get_simulator_docker_image(specifications)
        image_url = specifications['image']['url']

        cli = [cli] if cli else ['docker', 'run', '--tty', '--rm', image_url]

        result = subprocess.run(cli + ['-v'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        log = result.stdout.decode() if result.stdout else ''
        supported = re.search(r'\d+\.\d+', log)
        if not supported:
            warnings.warn(('Command-line interface should support the `-v` option for displaying version information inline.\n\n'
                           'The command-line interface displayed the following when executed with `-v`:\n\n  {}'
                           ).format(log.replace('\n', '\n  ')),
                          TestCaseWarning)

        result = subprocess.run(cli + ['--version'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        log = result.stdout.decode() if result.stdout else ''
        supported = re.search(r'\d+\.\d+', log)
        if not supported:
            warnings.warn(('Command-line interface should support the `--version` option for displaying version information inline.\n\n'
                           'The command-line interface displayed the following when executed with `--version`:\n\n  {}'
                           ).format(log.replace('\n', '\n  ')),
                          TestCaseWarning)



""" Common exceptions for test cases

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

__all__ = [
    'TestCaseException',
    'InvalidOutputsException',
    'SkippedTestCaseException',
    'TimeoutException',
]


class TestCaseException(Exception):
    """ Exception raised when outputs of execution of COMBINE/OMEX archive are not as expected """
    pass  # pragma: no cover


class InvalidOutputsException(TestCaseException):
    """ Exception raised when outputs of execution of COMBINE/OMEX archive are not as expected """
    pass  # pragma: no cover


class SkippedTestCaseException(TestCaseException):
    """ Exception raised that indicates that a test case should be skipped """
    pass  # pragma: no cover


class TimeoutException(TestCaseException):
    """ Exception raised that indicates that a test case timed out """
    pass  # pragma: no cover



""" Methods for test cases involving checking Docker images

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..data_model import TestCase
from ..warnings import TestCaseWarning
from .sedml import SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports
from biosimulators_utils.simulator.environ import ENVIRONMENT_VARIABLES
import warnings

__all__ = [
    'DefaultUserIsRoot',
    'DeclaresSupportedEnvironmentVariables',
    'HasOciLabels',
    'HasBioContainersLabels',
    'SingularityImageExecutesSimulationsSuccessfully',
]


class DefaultUserIsRoot(TestCase):
    """ Test that the default user of a Docker image is root """

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None, expected_user=(None, '', '0')):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            expected_user (:obj:`tuple`, optional): expected user
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        image = self.get_simulator_docker_image(specifications)
        user = image.attrs['Config']['User']
        if user not in expected_user:
            msg = ("The default user for the Docker image is `{}`. For compatability with Singularity, "
                   "Docker images for simulators should not declare default users (`USER`) other than root. "
                   "Compatability with Singularity is important for using the image on HPC systems, including "
                   "the infrastructure used by runBioSimulations and BioSimulations.\n\n"
                   "More information:\n"
                   "  https://biosimulators.org/conventions/simulator-images\n"
                   "  https://sylabs.io/guides/3.7/user-guide/singularity_and_docker.html#best-practices"
                   ).format(user)
            warnings.warn(msg, TestCaseWarning)


class DeclaresSupportedEnvironmentVariables(TestCase):
    """ Test if a Docker image declares the environment variables that is supports """

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        image = self.get_simulator_docker_image(specifications)
        expected_env_vars = set(var.name for var in ENVIRONMENT_VARIABLES.values())
        env_vars = set(key_val.partition('=')[0] for key_val in image.attrs['Config']['Env'])
        potentially_missing_env_vars = sorted(expected_env_vars.difference(env_vars))
        if potentially_missing_env_vars:
            msg = ('Docker images for simulation tools should declare the environment variables that they support.\n\n'
                   'The Docker image does not declare the following standard environment '
                   'variables recognized by BioSimulators:\n  - {}\n\n'
                   'If the simulation tool implements these variables, they should be declared in the Dockerfile for '
                   'the Docker image for the simulator.\n\n'
                   'Note, support for these environment variables is optional. Simulation tools are not required to support '
                   'these variables.'
                   ).format('\n  - '.join("'" + var + "'" for var in potentially_missing_env_vars))
            warnings.warn(msg, TestCaseWarning)


class HasOciLabels(TestCase):
    """ Test that a Docker image has Open Container Initiative (OCI) labels
    with metadata about the image
    """
    EXPECTED_LABELS = [
        'org.opencontainers.image.authors',
        'org.opencontainers.image.description',
        'org.opencontainers.image.documentation',
        'org.opencontainers.image.licenses',
        'org.opencontainers.image.revision',
        'org.opencontainers.image.source',
        'org.opencontainers.image.title',
        'org.opencontainers.image.url',
        'org.opencontainers.image.vendor',
        'org.opencontainers.image.version',
        'org.opencontainers.image.created',
    ]

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        image = self.get_simulator_docker_image(specifications)
        missing_labels = set(self.EXPECTED_LABELS).difference(set(image.labels.keys()))
        if missing_labels:
            warnings.warn('Docker images are encouraged to have the following Open Container Initiative (OCI) labels:\n  {}'.format(
                '\n  '.join(sorted(missing_labels))), TestCaseWarning)


class HasBioContainersLabels(TestCase):
    """ Test that a Docker image has BioContainers labels with metadata
    about the image
    """
    EXPECTED_LABELS = [
        "about.documentation",
        "about.home",
        "about.license",
        "about.license_file",
        "about.summary",
        "about.tags",
        "base_image",
        "extra.identifiers.biotools",
        "maintainer",
        "software",
        "software.version",
        "version",
    ]

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        image = self.get_simulator_docker_image(specifications)
        missing_labels = set(self.EXPECTED_LABELS).difference(set(image.labels.keys()))
        if missing_labels:
            warnings.warn('Docker images are encouraged to have the following BioContainers labels:\n  {}'.format(
                '\n  '.join(sorted(missing_labels))), TestCaseWarning)


class SingularityImageExecutesSimulationsSuccessfully(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that the Singularity version of a Docker image can sucessfully execute COMBINE archives """

    EXEC_WITH_SINGULARITY = True



""" Methods for test cases involving manually curated COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import Config
from ..data_model import (TestCase, SedTaskRequirements, ExpectedSedReport, ExpectedSedDataSet, ExpectedSedPlot,
                          AlertType, OutputMedium)
from ..exceptions import InvalidOutputsException, SkippedTestCaseException, TimeoutException, TestCaseException
from ..utils import get_singularity_image_filename, simulation_results_isnan
from ..warnings import IgnoredTestCaseWarning, SimulatorRuntimeErrorWarning, InvalidOutputsWarning
from .utils import are_array_shapes_equivalent
from biosimulators_utils.combine.data_model import CombineArchive, CombineArchiveContentFormatPattern  # noqa: F401
from biosimulators_utils.combine.io import CombineArchiveReader, CombineArchiveWriter
from biosimulators_utils.config import get_config
from biosimulators_utils.image import convert_docker_image_to_singularity
from biosimulators_utils.report.data_model import ReportFormat
from biosimulators_utils.report.io import ReportReader
from biosimulators_utils.sedml.data_model import (  # noqa: F401
    Report, Task, UniformTimeCourseSimulation,
    DataGenerator, Variable, Symbol, DataSet,
    Model, ModelLanguagePattern, Simulation, Algorithm)
from biosimulators_utils.sedml.io import SedmlSimulationReader, SedmlSimulationWriter
from biosimulators_utils.sedml.utils import (remove_algorithm_parameter_changes,
                                             replace_complex_data_generators_with_generators_for_individual_variables,
                                             remove_plots)
import biosimulators_utils.archive.io
import biosimulators_utils.simulator.exec
import biosimulators_utils.report.io
import abc
import glob
import json
import numpy
import numpy.testing
import os
import re
import shutil
import subprocess
import types  # noqa: F401
import warnings

__all__ = [
    'SimulatorCanExecutePublishedProject',
    'SyntheticCombineArchiveTestCase',
    'ExpectedResultOfSyntheticArchive',
    'find_cases',
    'ConfigurableMasterCombineArchiveTestCase',
    'SingleMasterSedDocumentCombineArchiveTestCase',
    'UniformTimeCourseTestCase',
]

EXAMPLES_DIR = os.path.join(os.path.dirname(__file__), '..', '..', 'examples')


class SimulatorCanExecutePublishedProject(TestCase):
    """ A test case for validating a simulator that involves executing a COMBINE/OMEX archive

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        filename (:obj:`str`): path to archive
        task_requirements (:obj:`list` of :obj:`SedTaskRequirements`): list of the required model format and simulation algorithm
            for each task in the COMBINE/OMEX archive
        skipped_simulators (:obj:`list` of :obj:`str`): list of simulation tools to not test the COMBINE archive with
        expected_reports (:obj:`list` of :obj:`ExpectedSedReport`): list of reports expected to be produced by algorithm
        expected_plots (:obj:`list` of :obj:`ExpectedSedPlot`): list of plots expected to be produced by algorithm
        runtime_failure_alert_type (:obj:`AlertType`): whether a run-time failure should be raised as an error or warning
        assert_no_extra_reports (:obj:`bool`): if :obj:`True`, raise an exception if the simulator produces unexpected reports
        assert_no_extra_datasets (:obj:`bool`): if :obj:`True`, raise an exception if the simulator produces unexpected datasets
        assert_no_missing_plots (:obj:`bool`): if :obj:`True`, raise an exception if the simulator doesn't produce the expected plots
        assert_no_extra_plots (:obj:`bool`): if :obj:`True`, raise an exception if the simulator produces unexpected plots
        r_tol (:obj:`float`): relative tolerence
        a_tol (:obj:`float`): absolute tolerence
    """

    def __init__(self, id=None, name=None, filename=None,
                 task_requirements=None, skipped_simulators=None,
                 expected_reports=None, expected_plots=None,
                 runtime_failure_alert_type=AlertType.exception,
                 assert_no_extra_reports=False, assert_no_extra_datasets=False,
                 assert_no_missing_plots=False, assert_no_extra_plots=False,
                 r_tol=1e-4, a_tol=0.,
                 output_medium=OutputMedium.console):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            filename (:obj:`str`, optional): path to archive
            task_requirements (:obj:`list` of :obj:`SedTaskRequirements`, optional): list of the required model format and simulation algorithm
                for each task in the COMBINE/OMEX archive
            skipped_simulators (:obj:`list` of :obj:`str`, optional): list of simulation tools to not test the COMBINE archive with
            expected_reports (:obj:`list` of :obj:`ExpectedSedReport`, optional): list of reports expected to be produced by algorithm
            expected_plots (:obj:`list` of :obj:`ExpectedSedPlot`, optional): list of plots expected to be produced by algorithm
            runtime_failure_alert_type (:obj:`AlertType`, optional): whether a run-time failure should be raised as an error or warning
            assert_no_extra_reports (:obj:`bool`, optional): if :obj:`True`, raise an exception if the simulator produces unexpected reports
            assert_no_extra_datasets (:obj:`bool`, optional): if :obj:`True`, raise an exception if the simulator produces unexpected datasets
            assert_no_missing_plots (:obj:`bool`, optional): if :obj:`True`, raise an exception if the simulator doesn't produce the expected plots
            assert_no_extra_plots (:obj:`bool`, optional): if :obj:`True`, raise an exception if the simulator produces unexpected plots
            r_tol (:obj:`float`, optional): relative tolerence
            a_tol (:obj:`float`, optional): absolute tolerence
            output_medium (:obj:`OutputMedium`, optional): medium the description should be formatted for
        """
        super(SimulatorCanExecutePublishedProject, self).__init__(id, name, output_medium=output_medium)
        self.filename = filename
        self.task_requirements = task_requirements or []
        self.skipped_simulators = skipped_simulators or []
        self.expected_reports = expected_reports or []
        self.expected_plots = expected_plots or []

        self.description = self.get_description()

        self.runtime_failure_alert_type = runtime_failure_alert_type
        self.assert_no_extra_reports = assert_no_extra_reports
        self.assert_no_extra_datasets = assert_no_extra_datasets
        self.assert_no_missing_plots = assert_no_missing_plots
        self.assert_no_extra_plots = assert_no_extra_plots
        self.r_tol = r_tol
        self.a_tol = a_tol

    def get_description(self):
        """ Get a description of the case

        Returns:
            :obj:`str`: description of the case
        """
        task_descriptions = []
        for req in self.task_requirements:
            format = '`{}`'.format(req.model_format)
            alg = '`{}`'.format(req.simulation_algorithm)

            if self.output_medium == OutputMedium.gh_issue:
                format_url = (
                    'https://www.ebi.ac.uk/ols/ontologies/edam/terms?iri='
                    + 'http%3A%2F%2Fedamontology.org%2F'
                    + req.model_format
                )
                alg_url = (
                    'https://www.ebi.ac.uk/ols/ontologies/kisao/terms?iri='
                    + 'http%3A%2F%2Fwww.biomodels.net%2Fkisao%2FKISAO%23'
                    + req.simulation_algorithm
                )
                format = '[{}]({})'.format(format, format_url)
                alg = '[{}]({})'.format(alg, alg_url)

            task_descriptions.append(''.join([
                '\n* Format: {}'.format(format),
                '\n  Algorithm: {}'.format(alg),
            ]))
        task_descriptions.sort()
        return 'Required model formats and simulation algorithms for SED tasks:\n' + ''.join(task_descriptions)

    def from_json(self, base_path, filename):
        """ Read test case from JSON file

        Args:
            base_path (:obj:`str`): bath directory for test cases
            filename (:obj:`str`): JSON file relative to :obj:`base_path`

        Returns:
            :obj:`SimulatorCanExecutePublishedProject`: this object
        """
        with open(os.path.join(base_path, filename), 'r') as file:
            data = json.load(file)

        id = filename.replace(os.sep + 'expected-results.json', '')
        self.id = 'published_project.SimulatorCanExecutePublishedProject' + ':' + id
        self.name = data['name']
        self.filename = os.path.join(
            base_path,
            os.path.relpath(os.path.join(os.path.dirname(filename), '..', data['filename']), '.'))

        return self.from_dict(data)

    def from_dict(self, data):
        """ Read test case from dictionary

        Args:
            data (:obj:`dict`): dictionary with test case data

        Returns:
            :obj:`SimulatorCanExecutePublishedProject`: this object
        """
        self.task_requirements = []
        for task_req_def in data['taskRequirements']:
            self.task_requirements.append(SedTaskRequirements(
                model_format=task_req_def['modelFormat'],
                model_format_features=set(task_req_def.get('modelFormatFeatures', [])),
                simulation_algorithm=task_req_def['simulationAlgorithm'],
            ))

        self.skipped_simulators = [simulator['id'] for simulator in data['skippedSimulators']]

        self.expected_reports = []
        for exp_report_def in data.get('expectedReports', []):
            id = exp_report_def['id']

            data_sets = [ExpectedSedDataSet(id=data_set.get('id', None), label=data_set.get('label', None))
                         for data_set in exp_report_def.get('dataSets', [])]
            data_set_ids = [data_set.id for data_set in data_sets]
            points = tuple(exp_report_def['points'])

            values = {}
            for labelVal in exp_report_def.get('values', []):
                data_set_id = labelVal['id']
                val = labelVal['value']
                if isinstance(val, dict):
                    values[data_set_id] = {}
                    for k, v in val.items():
                        multi_index = tuple(int(index) for index in k.split(","))
                        try:
                            numpy.ravel_multi_index([multi_index], points)[0]
                        except ValueError:
                            raise ValueError((
                                "Key `{}` of the expected values of report `{}` of published project test case `{}` is invalid. "
                                "Key must be less than or equal to `{}`."
                            ).format(
                                multi_index,
                                self.id,
                                self.id.replace('published_project.SimulatorCanExecutePublishedProject:', ''),
                                tuple(p - 1 for p in points),
                            ))
                        values[data_set_id][multi_index] = v
                else:
                    values[data_set_id] = numpy.array(val)

            invalid_dataset_ids = set(values.keys()).difference(set(data_set_ids))
            if invalid_dataset_ids:
                raise ValueError((
                    "The `id` fields of the expected values of report `{}` of published project test case `{}` "
                    "should be defined in the 'dataSets' property. "
                    "The following keys were not in the 'dataSets' property:\n  - {}").format(
                    id, self.id.replace('published_project.SimulatorCanExecutePublishedProject:', ''),
                    '\n  - '.join(sorted(invalid_dataset_ids))))

            self.expected_reports.append(ExpectedSedReport(
                id=id,
                data_sets=data_sets,
                points=points,
                values=values,
            ))

        self.expected_plots = []
        for exp_plot_def in data.get('expectedPlots', []):
            self.expected_plots.append(ExpectedSedPlot(
                id=exp_plot_def['id'],
            ))

        self.runtime_failure_alert_type = AlertType(data.get('runtimeFailureAlertType', 'exception'))
        self.assert_no_extra_reports = data.get('assertNoExtraReports', False)
        self.assert_no_extra_datasets = data.get('assertNoExtraDatasets', False)
        self.assert_no_missing_plots = data.get('assertNoMissingPlots', False)
        self.assert_no_extra_plots = data.get('assertNoExtraPlots', False)
        self.r_tol = data.get('r_tol', 1e-4)
        self.a_tol = data.get('a_tol', 0.)

        self.description = self.get_description()

        return self

    def compatible_with_specifications(self, specifications):
        # determine if case is applicable to simulator
        if specifications.get('id', None) in self.skipped_simulators:
            return False

        for task_reqs in self.task_requirements:
            reqs_satisfied = False
            for alg_specs in specifications['algorithms']:
                format_reqs_satisfied = False
                for format in alg_specs['modelFormats']:
                    if (
                        task_reqs.model_format == format['id']
                        and task_reqs.model_format_features == set(format.get('supportedFeatures', []) or [])
                    ):
                        format_reqs_satisfied = True
                        break
                if not format_reqs_satisfied:
                    break

                if task_reqs.simulation_algorithm == alg_specs['kisaoId']['id']:
                    reqs_satisfied = True
                    break

            if not reqs_satisfied:
                return False

        return True

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`SkippedTestCaseException`: if the test case is not applicable to the simulator
            :obj:`Exception`: if the simulator did not pass the test case
        """
        if not self.compatible_with_specifications(specifications):
            model_formats = set()
            simulation_algorithms = set()
            for task_req in self.task_requirements:
                model_formats.add(task_req.model_format)
                simulation_algorithms.add(task_req.simulation_algorithm)
            raise SkippedTestCaseException('Case requires model formats {} and simulation algorithms {}'.format(
                ', '.join(sorted(model_formats)), ', '.join(sorted(simulation_algorithms))))

        if dry_run:
            return

        if not os.path.isdir(working_dirname):
            os.makedirs(working_dirname)

        # pull image and execute COMBINE/OMEX archive for case
        try:
            self.exec_sedml_docs_in_archive(specifications, working_dirname, cli=cli)

        except Exception as exception:
            if os.path.isdir(working_dirname) and os.getenv('CI', 'false').lower() in ['1', 'true']:
                subprocess.run(['sudo', 'chown', '{}:{}'.format(os.getuid(), os.getgid()), '-R', working_dirname], check=True)

            if self.runtime_failure_alert_type == AlertType.exception:
                raise
            else:
                warnings.warn(str(exception), SimulatorRuntimeErrorWarning)
                return

        # check expected outputs created
        errors = []

        # check expected outputs created: reports
        if not os.path.isfile(os.path.join(working_dirname, get_config().H5_REPORTS_PATH)):
            errors.append('No reports were generated')

        else:
            report_reader = biosimulators_utils.report.io.ReportReader()
            for expected_report in self.expected_reports:
                report = Report()
                for data_set in expected_report.data_sets:
                    report.data_sets.append(DataSet(id=data_set.id, label=data_set.label))
                try:
                    report_results = report_reader.run(report, working_dirname, expected_report.id, format=ReportFormat.h5)
                except Exception:
                    errors.append('Report {} could not be read'.format(expected_report.id))
                    continue

                missing_data_sets = set([data_set.id for data_set in expected_report.data_sets]).difference(set(report_results.keys()))
                if missing_data_sets:
                    errors.append(('Report {} does not contain expected data sets:\n  {}\n\n'
                                   'Report contained these data sets:\n  {}').format(
                        expected_report.id,
                        '\n  '.join(sorted(missing_data_sets)),
                        '\n  '.join(sorted(report_results.keys())),
                    ))
                    continue

                points = report_results[report.data_sets[0].id].shape
                if not are_array_shapes_equivalent(points, expected_report.points):
                    errors.append('Report {} contains incorrect number of points: {} != {}'.format(
                                  expected_report.id, points, expected_report.points))
                    continue

                for data_set_id, expected_value in expected_report.values.items():
                    if isinstance(expected_value, dict):
                        value = report_results[data_set_id]
                        for el_id, expected_el_value in expected_value.items():
                            el_index = numpy.ravel_multi_index([el_id], value.shape)[0]
                            actual_el_value = value[el_index]
                            try:
                                numpy.testing.assert_allclose(
                                    actual_el_value,
                                    expected_el_value,
                                    rtol=self.r_tol,
                                    atol=self.a_tol,
                                )
                            except AssertionError:
                                errors.append('Data set {} of report {} does not have expected value at {}: {} != {}'.format(
                                    data_set_id, expected_report.id, el_id, actual_el_value, expected_el_value))
                    else:
                        try:
                            numpy.testing.assert_allclose(
                                report_results[data_set_id],
                                expected_value,
                                rtol=self.r_tol,
                                atol=self.a_tol,
                            )
                        except AssertionError:
                            errors.append('Data set {} of report {} does not have expected values'.format(
                                data_set_id, expected_report.id))

            report_ids = set(report_reader.get_ids(working_dirname))
            expected_report_ids = set(report.id for report in self.expected_reports)
            extra_report_ids = report_ids.difference(expected_report_ids)
            if extra_report_ids:
                if self.assert_no_extra_reports:
                    errors.append('Unexpected reports were produced:\n  {}'.format(
                        '\n  '.join(sorted(extra_report_ids))))
                else:
                    warnings.warn('Unexpected reports were produced:\n  {}'.format(
                        '\n  '.join(sorted(extra_report_ids))), InvalidOutputsWarning)

        # check expected outputs created: plots
        if os.path.isfile(os.path.join(working_dirname, get_config().PLOTS_PATH)):
            archive = biosimulators_utils.archive.io.ArchiveReader().run(os.path.join(working_dirname, 'plots.zip'))
            plot_ids = set(os.path.splitext(file.archive_path)[0] for file in archive.files)
        else:
            plot_ids = set()

        expected_plot_ids = set(plot.id for plot in self.expected_plots)

        missing_plot_ids = expected_plot_ids.difference(plot_ids)
        extra_plot_ids = plot_ids.difference(expected_plot_ids)

        if missing_plot_ids:
            if self.assert_no_missing_plots:
                errors.append('Plots were not produced:\n  {}'.format(
                    '\n  '.join(sorted(missing_plot_ids))))
            else:
                warnings.warn('Plots were not produced:\n  {}'.format(
                    '\n  '.join(sorted(missing_plot_ids))), InvalidOutputsWarning)

        if extra_plot_ids:
            if self.assert_no_extra_plots:
                errors.append('Extra plots were produced:\n  {}'.format(
                    '\n  '.join(sorted(extra_plot_ids))))
            else:
                warnings.warn('Extra plots were produced:\n  {}'.format(
                    '\n  '.join(sorted(extra_plot_ids))), InvalidOutputsWarning)

        # raise errors
        if errors:
            raise InvalidOutputsException('\n\n'.join(errors))

    def exec_sedml_docs_in_archive(self, specifications, out_dir, cli=None):
        """
        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            out_dir (:obj:`str`): path to save simulation results
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image
        """
        config = Config()
        pull_docker_image = config.pull_docker_image
        user_to_exec_within_container = config.user_to_exec_in_simulator_containers
        if os.getenv('CI', 'false').lower() in ['1', 'true']:
            user_to_exec_within_container = '_SUDO_'

        if cli:
            biosimulators_utils.simulator.exec.exec_sedml_docs_in_archive_with_simulator_cli(
                self.filename, out_dir, cli)

        else:
            biosimulators_utils.simulator.exec.exec_sedml_docs_in_archive_with_containerized_simulator(
                self.filename, out_dir, specifications['image']['url'], pull_docker_image=pull_docker_image,
                user_to_exec_within_container=user_to_exec_within_container)

            if os.path.isdir(out_dir) and os.getenv('CI', 'false').lower() in ['1', 'true']:
                subprocess.run(['sudo', 'chown', '{}:{}'.format(os.getuid(), os.getgid()), '-R', out_dir], check=True)


class SyntheticCombineArchiveTestCase(TestCase):
    """ Test that involves a computationally-generated COMBINE/OMEX archive

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        description (:obj:`str`): description
        output_medium (:obj:`OutputMedium`): medium the description should be formatted for
        published_projects_test_cases (:obj:`list` of :obj:`SimulatorCanExecutePublishedProject`):
            curated COMBINE/OMEX archives that can be used to generate example archives for testing
        _published_projects_test_case (:obj:`SimulatorCanExecutePublishedProject`): COMBINE/OMEX archive
            that is used to generate example archives for testing
    """

    EXEC_WITH_SINGULARITY = False
    REPORT_ERROR_AS_SKIP = False

    def __init__(self, id=None, name=None, description=None, output_medium=OutputMedium.console, published_projects_test_cases=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            description (:obj:`str`): description
            output_medium (:obj:`OutputMedium`, optional): medium the description should be formatted for
            published_projects_test_cases (:obj:`list` of :obj:`SimulatorCanExecutePublishedProject`, optional):
                curated COMBINE/OMEX archives that can be used to generate example archives for testing
        """
        super(SyntheticCombineArchiveTestCase, self).__init__(id=id, name=name, description=description, output_medium=output_medium)
        self.published_projects_test_cases = published_projects_test_cases or []
        self._published_projects_test_case = None

    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        try:
            return_value = self._eval(specifications, working_dirname,
                                      synthetic_archives_dir=synthetic_archives_dir, dry_run=dry_run, cli=cli)
        except Exception as exception:
            if not isinstance(exception, TimeoutException) and self.REPORT_ERROR_AS_SKIP:
                raise SkippedTestCaseException(str(exception))
            else:
                raise

        return return_value

    def _eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs

        Raises:
            :obj:`Exception`: if the simulator did not pass the test case
        """
        if not os.path.isdir(working_dirname):
            os.makedirs(working_dirname)

        # read curated archives and find one that is suitable for testing
        suitable_curated_archive = False
        for published_projects_test_case in self.published_projects_test_cases:
            self._published_projects_test_case = published_projects_test_case

            # read archive
            curated_archive_filename = published_projects_test_case.filename
            shared_archive_dir = os.path.join(working_dirname, 'archive')
            if not os.path.isdir(shared_archive_dir):
                os.makedirs(shared_archive_dir)

            curated_archive = CombineArchiveReader().run(curated_archive_filename, shared_archive_dir)
            curated_sed_docs = {}
            sedml_reader = SedmlSimulationReader()
            for content in list(curated_archive.contents):
                if content.format and re.match(CombineArchiveContentFormatPattern.SED_ML, content.format):
                    sed_doc = sedml_reader.run(os.path.join(shared_archive_dir, content.location))
                    curated_sed_docs[content.location] = sed_doc

                # remove manifest from contents because libSED-ML occassionally has trouble with this
                elif (
                    content.location in ['manifest.xml', './manifest.xml']
                    and content.format == 'http://identifiers.org/combine.specifications/omex-manifest'
                ):
                    curated_archive.contents.remove(content)
                    os.remove(os.path.join(shared_archive_dir, content.location))

            # see if archive is suitable for testing
            if self.is_curated_archive_suitable_for_building_synthetic_archive(specifications, curated_archive, curated_sed_docs):
                suitable_curated_archive = True
                break

            # cleanup
            shutil.rmtree(shared_archive_dir)

        if not suitable_curated_archive:
            raise SkippedTestCaseException('No curated COMBINE/OMEX archives are available to generate archives for testing')

        expected_results_of_synthetic_archives = self.build_synthetic_archives(
            specifications, curated_archive, shared_archive_dir, curated_sed_docs)
        has_warnings = False
        for i_archive, expected_results_of_synthetic_archive in enumerate(expected_results_of_synthetic_archives):
            if self._eval_synthetic_archive(specifications, expected_results_of_synthetic_archive, shared_archive_dir,
                                            i_archive, os.path.join(working_dirname, str(i_archive + 1)),
                                            synthetic_archives_dir=synthetic_archives_dir, dry_run=dry_run,
                                            cli=cli):
                has_warnings = True
        return not has_warnings

    def _eval_synthetic_archive(self, specifications, expected_results_of_synthetic_archive, shared_archive_dir,
                                i_synthetic_archive, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        synthetic_archive = expected_results_of_synthetic_archive.archive
        synthetic_sed_docs = expected_results_of_synthetic_archive.sed_documents
        is_success_expected = expected_results_of_synthetic_archive.is_success_expected
        environment = expected_results_of_synthetic_archive.environment

        if not os.path.isdir(working_dirname):
            os.makedirs(working_dirname)

        sedml_writer = SedmlSimulationWriter()
        synthetic_archive_filename = os.path.join(working_dirname, 'archive.omex')
        for location, sed_doc in synthetic_sed_docs.items():
            sedml_writer.run(sed_doc, os.path.join(shared_archive_dir, location))
        CombineArchiveWriter().run(synthetic_archive, shared_archive_dir, synthetic_archive_filename)

        if synthetic_archives_dir:
            cls = self.__class__
            module = cls.__module__.partition('biosimulators_test_suite.test_case.')[2]
            export_synthetic_archive_dirname = os.path.join(synthetic_archives_dir, module, cls.__name__)
            if not os.path.isdir(export_synthetic_archive_dirname):
                os.makedirs(export_synthetic_archive_dirname)
            export_synthetic_archive_filename = os.path.join(export_synthetic_archive_dirname,
                                                             '{}.{}.omex'.format(
                                                                 str(i_synthetic_archive + 1),
                                                                 'execution-should-succeed'
                                                                 if is_success_expected else
                                                                 'execute-should-fail'))
            shutil.copy(synthetic_archive_filename, export_synthetic_archive_filename)

        if dry_run:
            return False

        # use synthetic archive to test simulator
        outputs_dir = os.path.join(working_dirname, 'outputs')
        config = Config()
        pull_docker_image = config.pull_docker_image
        user_to_exec_within_container = config.user_to_exec_in_simulator_containers
        has_warnings = False
        try:
            if os.getenv('CI', 'false').lower() in ['1', 'true']:
                user_to_exec_within_container = '_SUDO_'

            if cli:
                biosimulators_utils.simulator.exec.exec_sedml_docs_in_archive_with_simulator_cli(
                    synthetic_archive_filename, outputs_dir, cli, environment=environment)

            elif self.EXEC_WITH_SINGULARITY:
                docker_image_url = specifications['image']['url']

                # get path for Singularity image
                singularity_filename = get_singularity_image_filename(docker_image_url)

                # convert image to Singularity format
                convert_docker_image_to_singularity(docker_image_url, singularity_filename=singularity_filename)

                # run a simulation with the Singularity image
                if not os.path.isdir(outputs_dir):
                    os.makedirs(outputs_dir)
                temp_filename = os.path.join(outputs_dir, os.path.basename(synthetic_archive_filename))
                shutil.copyfile(synthetic_archive_filename, temp_filename)

                cmd = [
                    'singularity', 'run',
                    '-B', outputs_dir + ':/root',
                    singularity_filename,
                    '-i', '/root/' + os.path.basename(synthetic_archive_filename),
                    '-o', '/root',
                ]
                result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
                os.remove(temp_filename)
                if result.returncode != 0:
                    msg = 'The Docker image could not be successfully executed as a Singularity image:\n  {}'.format(
                        result.stderr.decode().replace('\n', '\n  '))
                    raise TestCaseException(msg)

            else:
                biosimulators_utils.simulator.exec.exec_sedml_docs_in_archive_with_containerized_simulator(
                    synthetic_archive_filename, outputs_dir, specifications['image']['url'], pull_docker_image=pull_docker_image,
                    environment=environment,
                    user_to_exec_within_container=user_to_exec_within_container)

            if os.path.isdir(outputs_dir) and os.getenv('CI', 'false').lower() in ['1', 'true']:
                subprocess.run(['sudo', 'chown', '{}:{}'.format(os.getuid(), os.getgid()), '-R', outputs_dir], check=True)

            if not self.eval_outputs(specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
                has_warnings = True

            succeeded = True

        except Exception:
            if os.path.isdir(outputs_dir) and os.getenv('CI', 'false').lower() in ['1', 'true']:
                subprocess.run(['sudo', 'chown', '{}:{}'.format(os.getuid(), os.getgid()), '-R', outputs_dir], check=True)

            succeeded = False
            if is_success_expected:
                raise

        if succeeded and not is_success_expected:
            msg = 'The execution of the COMBINE/OMEX archive did not fail as expected'
            raise ValueError(msg)

        return has_warnings

    def is_curated_archive_suitable_for_building_synthetic_archive(self, specifications, archive, sed_docs):
        """ Find an archive with at least one report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`bool`: :obj:`True`, if the curated archive is suitable for generating a synthetic
                archive for testing
        """
        for location, sed_doc in sed_docs.items():
            if self.is_curated_sed_doc_suitable_for_building_synthetic_archive(specifications, sed_doc, location):
                return True
        return False

    def is_curated_sed_doc_suitable_for_building_synthetic_archive(self, specifications, sed_doc, sed_doc_location):
        """ Determine if a SED document is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            sed_doc (:obj:`SedDocument`): SED document in curated archive
            sed_doc_location (:obj:`str`): location of the SED document within its parent COMBINE/OMEX archive

        Returns:
            :obj:`bool`: whether the SED document is suitable for testing
        """
        split_sed_doc_location = os.path.split(os.path.relpath(sed_doc_location, '.'))
        if split_sed_doc_location[0] or len(split_sed_doc_location) > 2:
            return False

        for output in sed_doc.outputs:
            if isinstance(output, Report) and self.is_curated_sed_report_suitable_for_building_synthetic_archive(
                    specifications, output, sed_doc_location):
                return True

        return False

    def is_curated_sed_report_suitable_for_building_synthetic_archive(self, specifications, report, sed_doc_location):
        """ Determine if a SED report is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            report (:obj:`Report`): SED report in curated archive
            sed_doc_location (:obj:`str`): location of the SED document within its parent COMBINE/OMEX archive

        Returns:
            :obj:`bool`: whether the report is suitable for testing
        """
        task_variables = {}
        for data_set in report.data_sets:
            for variable in data_set.data_generator.variables:
                if variable.task not in task_variables:
                    task_variables[variable.task] = set()
                task_variables[variable.task].add(variable.id)

        for task, variables in task_variables.items():
            if len(variables) >= 1 and self.is_curated_sed_task_suitable_for_building_synthetic_archive(specifications, task):
                return True

        return False

    def is_curated_sed_task_suitable_for_building_synthetic_archive(self, specifications, task):
        """ Determine if a SED task is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            task (:obj:`Task`): SED task in curated archive

        Returns:
            :obj:`bool`: whether the task is suitable for testing
        """
        if not isinstance(task, Task):
            return False

        if not self.is_curated_sed_model_suitable_for_building_synthetic_archive(specifications, task.model):
            return False

        if not self.is_curated_sed_simulation_suitable_for_building_synthetic_archive(specifications, task.simulation):
            return False

        return True

    def is_curated_sed_model_suitable_for_building_synthetic_archive(self, specifications, model):
        """ Determine if a SED model is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            model (:obj:`Model`): SED model in curated archive

        Returns:
            :obj:`bool`: whether the model is suitable for testing
        """
        if not model or not model.source:
            return False
        source = model.source.lower()
        return not (
            source.startswith('#')
            or source.startswith('http://')
            or source.startswith('https://')
            or source.startswith('urn:')
        )

    def is_curated_sed_simulation_suitable_for_building_synthetic_archive(self, specifications, simulation):
        """ Determine if a SED simulation is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            simulation (:obj:`Simulation`): SED simulation in curated archive

        Returns:
            :obj:`bool`: whether the simulation is suitable for testing
        """
        if isinstance(simulation, UniformTimeCourseSimulation):
            if (
                simulation.initial_time != 0
                or simulation.output_start_time != 0
                or simulation.number_of_points < 1
                or int(simulation.number_of_points / 2) != simulation.number_of_points / 2
            ):
                return False
        return self.is_curated_sed_algorithm_suitable_for_building_synthetic_archive(specifications, simulation.algorithm)

    def is_curated_sed_algorithm_suitable_for_building_synthetic_archive(self, specifications, algorithm):
        """ Determine if a SED algorithm is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            algorithm (:obj:`Algorithm`): SED algorithm in curated archive

        Returns:
            :obj:`bool`: whether the algorithm is suitable for testing
        """
        return True

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        return [ExpectedResultOfSyntheticArchive(curated_archive, curated_sed_docs, True)]

    @abc.abstractmethod
    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        pass  # pragma: no cover


class ExpectedResultOfSyntheticArchive(object):
    """ An expected result of executing a synthetic COMBINE/OMEX archive

    Attributes:
        archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
        sed_documents (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
          SED documents in synthetic archive
        is_success_expected (:obj:`bool`, optional): whether the execution of the archive is expected to succeed
        environment (:obj:`dict`, optional): environment variables that the archive should be executed with
    """

    def __init__(self, archive, sed_documents, is_success_expected=True, environment=None):
        self.archive = archive
        self.sed_documents = sed_documents
        self.is_success_expected = is_success_expected
        self.environment = environment or {}


def find_cases(specifications, dir_name=None, output_medium=OutputMedium.console):
    """ Collect test cases

    Args:
        specifications (:obj:`dict`): specifications of the simulator to validate
        dir_name (:obj:`str`, optional): path to find example COMBINE/OMEX archives
        output_medium (:obj:`OutputMedium`, optional): medium the description should be formatted for

    Returns:
        :obj:`list` of :obj:`SimulatorCanExecutePublishedProject`: test cases
    """
    if dir_name is None:
        dir_name = EXAMPLES_DIR
    if not os.path.isdir(dir_name):
        warnings.warn('Directory of example COMBINE/OMEX archives is not available', IgnoredTestCaseWarning)

    all_cases = []
    compatible_cases = []
    for example_filename in glob.glob(os.path.join(dir_name, '**/*.omex'), recursive=True):
        md_filename = os.path.join(example_filename[0:-5], 'expected-results.json')
        rel_filename = os.path.relpath(md_filename, dir_name)
        case = SimulatorCanExecutePublishedProject(output_medium=output_medium).from_json(dir_name, rel_filename)
        all_cases.append(case)
        if case.compatible_with_specifications(specifications):
            compatible_cases.append(case)

    all_cases.sort(key=lambda case: case.filename)
    compatible_cases.sort(key=lambda case: case.filename)

    # return cases
    return (all_cases, compatible_cases)


class ConfigurableMasterCombineArchiveTestCase(SyntheticCombineArchiveTestCase):
    """ Class for generating synthetic archives with a single master SED-ML file or two non-master
    copies of the same file

    Attributes:
        _archive_has_master (:obj:`bool`): whether the synthetic archive should have a master file
        _model_change_filter (:obj:`types.FunctionType`): filter for model changes to keep
        _remove_algorithm_parameter_changes (:obj:`bool`): if :obj:`True`, remove instructions to change
            the values of the parameters of algorithms
        _use_single_variable_data_generators (:obj:`bool`): if :obj:`True`, replace data generators that
            involve multiple variables or parameters (and data sets, curves, and surfaces) with multiple
            data generators for each variable
        _remove_plots (:obj:`bool`): if :obj:`True`, remove plots
        _keep_one_task_one_report (:obj:`bool`): if :obj:`True`, keep only 1 task and only 1 report
        _expected_report_ids (:obj:`list` of :obj:`str`): ids of expected reports
    """

    @property
    @abc.abstractmethod
    def _archive_has_master(self):
        pass  # pragma: no cover

    def __init__(self, *args, **kwargs):
        super(ConfigurableMasterCombineArchiveTestCase, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: False
        self._remove_algorithm_parameter_changes = True
        self._use_single_variable_data_generators = True
        self._remove_plots = True
        self._keep_one_task_one_report = True

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with master and non-master SED documents

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        super(ConfigurableMasterCombineArchiveTestCase, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)

        # get a suitable SED document to modify
        for doc_location, doc in curated_sed_docs.items():
            if self.is_curated_sed_doc_suitable_for_building_synthetic_archive(specifications, doc, doc_location):
                break
        doc_content = next(content for content in curated_archive.contents if content.location == doc_location)

        # remove all other SED documents from archive
        original_contents = curated_archive.contents
        curated_archive.contents = []
        for content in original_contents:
            if content == doc_content or content.location not in curated_sed_docs:
                curated_archive.contents.append(content)

        # make all content not master
        for content in curated_archive.contents:
            content.master = False

        # make document master
        doc_content.master = self._archive_has_master

        # retain some model changes
        for model in doc.models:
            model.changes = list(filter(self._model_change_filter, model.changes))

        # remove algorithm parameter changes
        if self._remove_algorithm_parameter_changes:
            remove_algorithm_parameter_changes(doc)

        # replace data generator that involve mathematical expressions with multiple data generators for each individual variable
        if self._use_single_variable_data_generators:
            replace_complex_data_generators_with_generators_for_individual_variables(doc)

        # remove plots
        if self._remove_plots:
            remove_plots(doc)

        # keep only single task and single report
        if self._keep_one_task_one_report and self._use_single_variable_data_generators and self._remove_plots:
            key_report = None
            key_task = None
            for report in doc.outputs:
                if self.is_curated_sed_report_suitable_for_building_synthetic_archive(specifications, report, doc_location):
                    for data_set in report.data_sets:
                        task = data_set.data_generator.variables[0].task
                        if self.is_curated_sed_task_suitable_for_building_synthetic_archive(specifications, task):
                            key_report = report
                            key_task = task
                            break
                    if key_report:
                        break

            key_sim = key_task.simulation
            if isinstance(key_sim, UniformTimeCourseSimulation):
                key_sim.output_end_time = (
                    key_sim.output_start_time
                    + 10 / key_sim.number_of_points * (key_sim.output_end_time - key_sim.output_start_time)
                )
                key_sim.number_of_points = 10

            doc.models = [key_task.model]
            doc.simulations = [key_task.simulation]
            doc.tasks = [key_task]
            doc.data_generators = [data_gen for data_gen in doc.data_generators if data_gen.variables[0].task == key_task]
            doc.outputs = [key_report]
            key_report.data_sets = [data_set for data_set in key_report.data_sets if data_set.data_generator in doc.data_generators]

            # limit number of data set
            max_data_sets = 10
            key_report.data_sets = key_report.data_sets[0:max_data_sets]
            doc.data_generators = [data_set.data_generator for data_set in key_report.data_sets]

        curated_sed_docs = {
            doc_content.location: doc,
        }

        self._expected_report_ids = []
        for output in doc.outputs:
            if isinstance(output, Report):
                self._expected_report_ids.append(os.path.join(os.path.relpath(doc_content.location, './'), output.id))

        # return modified SED document
        return [ExpectedResultOfSyntheticArchive(curated_archive, curated_sed_docs, True)]


class SingleMasterSedDocumentCombineArchiveTestCase(ConfigurableMasterCombineArchiveTestCase):
    """ Class for generating synthetic COMBINE/OMEX archives with a single master SED-ML file

    Attributes:
        _archive_has_master (:obj:`bool`): whether the synthetic archive should  have a master file
    """

    @property
    def _archive_has_master(self):
        return True


class UniformTimeCourseTestCase(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that a simulator supports multiple reports per SED document """

    TEST_TIME = True

    def is_curated_sed_task_suitable_for_building_synthetic_archive(self, specifications, task):
        """ Determine if a SED task is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            task (:obj:`Task`): SED task in curated archive

        Returns:
            :obj:`bool`: whether the task is suitable for testing
        """
        if not super(UniformTimeCourseTestCase, self) \
                .is_curated_sed_task_suitable_for_building_synthetic_archive(specifications, task):
            return False

        sim = task.simulation
        return (
            isinstance(sim, UniformTimeCourseSimulation)
            and sim.initial_time == 0
            and sim.output_start_time == 0
            and int(sim.number_of_points / 2) == sim.number_of_points / 2
        )

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(UniformTimeCourseTestCase, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)

        for expected_results_of_synthetic_archive in expected_results_of_synthetic_archives:
            curated_archive = expected_results_of_synthetic_archive.archive
            curated_sed_docs = expected_results_of_synthetic_archive.sed_documents

            # get a suitable SED document to modify
            doc = list(curated_sed_docs.values())[0]
            task = doc.tasks[0]
            sim = task.simulation
            self.modify_simulation(sim)
            report = doc.outputs[0]

            if self.TEST_TIME:
                self.add_time_data_set(doc, task, report)

            expected_results_of_synthetic_archive.archive = curated_archive
            expected_results_of_synthetic_archive.sed_documents = curated_sed_docs

        # return modified SED document
        return expected_results_of_synthetic_archives

    def add_time_data_set(self, doc, task, report):
        """ Rename or add a time data set to a SED report

        Args:
            doc (:obj:`SedDocument`): SED document
            task (:obj:`task`): SED task
            report (:obj:`Report`): SED report
        """
        time_data_gen = None
        for data_gen in doc.data_generators:
            var = data_gen.variables[0]
            if var.task == task and var.symbol == Symbol.time:
                time_data_gen = data_gen
                break

        if not time_data_gen:
            if re.match(ModelLanguagePattern.CellML.value, task.model.language):
                msg = (
                    'This test case requires a model language which supports the time symbol ({}). '
                    '{} does not support the time symbol.'
                ).format(Symbol.time.value, ModelLanguagePattern.CellML.name)
                raise SkippedTestCaseException(msg)

            time_data_gen = DataGenerator(
                id='__data_generator_time__',
                variables=[
                    Variable(
                        id='__variable_time__',
                        task=task,
                        symbol=Symbol.time,
                    ),
                ],
                math='__variable_time__',
            )
            doc.data_generators.append(time_data_gen)

        time_data_set = None
        for data_set in report.data_sets:
            if data_set.data_generator == time_data_gen:
                time_data_set = data_set
                data_set.id = '__data_set_time__'
                break

        if not time_data_set:
            report.data_sets.append(
                DataSet(
                    id='__data_set_time__',
                    label='__data_set_time__',
                    data_generator=time_data_gen,
                )
            )

    @abc.abstractmethod
    def modify_simulation(self, simulation):
        """ Modify a simulation

        Args:
            simulation (:obj:`UniformTimeCourseSimulation`): simulation
        """
        pass  # pragma: no cover

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        has_warnings = False

        doc_location = list(synthetic_sed_docs.keys())[0]
        doc_id = os.path.relpath(doc_location, './')
        doc = list(synthetic_sed_docs.values())[0]
        sim = doc.simulations[0]
        report = doc.outputs[0]

        data = ReportReader().run(report, outputs_dir, os.path.join(doc_id, report.id))

        for data_set_data in data.values():
            if numpy.any(simulation_results_isnan(data_set_data)):
                warnings.warn('The results produced by the simulator include `NaN`.', InvalidOutputsWarning)
                has_warnings = True
                break

        if self.TEST_TIME:
            try:
                numpy.testing.assert_allclose(
                    data['__data_set_time__'],
                    numpy.linspace(sim.output_start_time, sim.output_end_time, sim.number_of_points + 1),
                    rtol=1e-4,
                )
            except Exception as exception:
                raise ValueError('Simulator did not produce the expected time course:\n\n  {}'.format(
                    str(exception).replace('\n', '\n  ')))

        return not has_warnings



""" Methods for processing submissions to the BioSimulators registry
(CI workflows for reviewing and committing simulators to the registry)

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .config import Config
from .data_model import OutputMedium
from .exec_core import SimulatorValidator
from .results.data_model import TestCaseResult, TestCaseResultType, TestResultsReport  # noqa: F401
from .results.io import write_test_results
from .utils import get_singularity_image_filename
from biosimulators_utils.biosimulations.utils import validate_biosimulations_api_response
from biosimulators_utils.config import Colors
from biosimulators_utils.gh_action.data_model import Comment, GitHubActionCaughtError  # noqa: F401
from biosimulators_utils.gh_action.core import GitHubAction, GitHubActionErrorHandling
from biosimulators_utils.image import get_docker_image
from biosimulators_utils.simulator_registry.data_model import SimulatorSubmission, IssueLabel  # noqa: F401
from biosimulators_utils.simulator_registry.process_submission import get_simulator_submission_from_gh_issue_body
from biosimulators_utils.simulator_registry.query import get_simulator_version_specs
from natsort import natsort_keygen
import biosimulators_utils.image
import biosimulators_utils.simulator.io
import requests
import requests.exceptions
import termcolor


__all__ = [
    'ValidateCommitSimulatorGitHubAction',
]


def get_uncaught_exception_msg(exception):
    """ Create an error message to display to users for all exceptions not caught during the
    exception of the :obj:`run` method (exceptions of all types except :obj:`GitHubActionCaughtError`)

    Args:
        exception (:obj:`Exception`): a failure encountered during the exception of the :obj:`run` method

    Returns:
        :obj:`str`: error message to display to users
    """
    gh_action_run_url = GitHubAction.get_gh_action_run_url()
    return [
        Comment(text='The validation/submission of your simulator failed.'),
        Comment(text=str(exception), error=True),
        Comment(text=('The complete log of your validation/submission job, including further information about the failure, '
                      + 'is available for 90 days [here]({}).'.format(gh_action_run_url))),
        Comment(text=('If you chose to validate your Docker image, the results of the validation of your image will be '
                      'available shortly as a JSON file. A link to this file will be available for 90 days from the "Artifacts" '
                      'section at the bottom of [this page]({}).'.format(gh_action_run_url))),
        Comment(text='Once you have fixed the problem, edit the first block of this issue to re-initiate this validation.'),
        Comment(text=('The BioSimulators Team is happy to help. '
                      'Questions and feedback can be directed to the BioSimulators Team by posting comments to this issues that '
                      'reference the GitHub team `@biosimulators/biosimulators` (without the backticks).')),
    ]


class ValidateCommitSimulatorGitHubAction(GitHubAction):
    """ Action to validate a containerized simulator

    Attributes:
        issue_number (:obj:`str`): number of GitHub issue which triggered the action
    """

    def __init__(self):
        super(ValidateCommitSimulatorGitHubAction, self).__init__()
        self.config = Config()
        self.issue_number = self.get_issue_number()

    @GitHubActionErrorHandling.catch_errors(uncaught_exception_msg_func=get_uncaught_exception_msg,
                                            caught_error_labels=[IssueLabel.invalid],
                                            uncaught_error_labels=[IssueLabel.invalid])
    def run(self):
        """ Validate and commit a simulator."""

        # Get properties of submission
        issue_props = self.get_issue(self.issue_number)
        submission = get_simulator_submission_from_gh_issue_body(issue_props['body'])
        submitter = issue_props['user']['login']

        # Send message that validation is starting
        self.add_comment_to_issue(self.issue_number, self.get_initial_message(submission, submitter))

        # reset labels except approved
        self.reset_issue_labels(self.issue_number, [IssueLabel.validated.value, IssueLabel.invalid.value, IssueLabel.action_error.value])

        # get specifications of simulator and validate simulator
        specifications, test_results = self.exec_core(submission, submitter)

        # label issue as validated
        self.add_labels_to_issue(self.issue_number, [IssueLabel.validated.value])

        # get specifications of other versions of simulator
        existing_version_specifications = get_simulator_version_specs(specifications['id'])

        # determine if simulator is approved: issue is a revision of an existing version of a validated simulator,
        # a new version of a validated simulator, or the issue has been manually approved by the BioSimulators Team
        approved = self.is_simulator_approved(specifications, existing_version_specifications)

        # if approved, label the issue as approved
        if approved and IssueLabel.approved.value not in self.get_labels_for_issue(self.issue_number):
            self.add_labels_to_issue(self.issue_number, [IssueLabel.approved.value])

        # commit simulator or indicate that further review is required
        if submission.commit_simulator:
            if approved:
                self.commit_simulator(submission, specifications, existing_version_specifications, test_results)

                # post success message
                self.add_comment_to_issue(
                    self.issue_number,
                    ''.join([
                        'Your submission was committed to the BioSimulators registry. Thank you!\n',
                        '\n',
                        'Future submissions of subsequent versions of {} to the BioSimulators registry '.format(specifications['id']),
                        'will be automatically validated. These submissions will not require manual review by the BioSimulators Team.',
                    ])
                )

                # close issue
                self.close_issue(self.issue_number)
            else:
                assigned = set([assignee['login'] for assignee in issue_props['assignees']])
                new_assignees = set(self.config.biosimulators_curator_gh_ids).difference(assigned)
                if new_assignees:
                    self.assign_issue(self.issue_number,  list(new_assignees))
                self.add_comment_to_issue(
                    self.issue_number,
                    ('A member of the BioSimulators team will review your submission and '
                     'publish your image before final committment to the registry.'))

    def get_initial_message(self, submission, submitter):
        """ Peport message that validation is starting

        Args:
            submission (:obj:`SimulatorSubmission`): simulator submission
            submitter (:obj:`str`): GitHub user name of person who executed the submission
        """
        actions = []
        not_actions = []

        actions.append('validating the specifications of your simulator')
        if submission.validate_image:
            actions.append('validating your Docker image')
            test_results_msg = (
                ' The results of the validation of your tool will also be saved as a JSON file. '
                'A link to this file will be available for 90 days from the "Artifacts" section at the bottom of this page.'
            )
        else:
            not_actions.append('You have chosen not to have the Docker image for your simulator validated.')
            test_results_msg = ''
        if submission.commit_simulator:
            job_type = 'submission'
            actions.append('committing your simulator to the BioSimulators registry')
        else:
            job_type = 'validation'
            not_actions.append('You have chosen not to submit your simulator to the BioSimulators registry.')

        if len(actions) == 1:
            actions = actions[0]
        else:
            actions = ', '.join(actions[0:-1]) + ', and ' + actions[-1]
        not_actions = ' '.join(action.strip() for action in not_actions)

        return ('Thank you @{} for your submission to the BioSimulators simulator validation/submission system!\n\n'
                'The BioSimulators validator bot is {}. {}\n\n'
                'We will discuss any concerns with your submission in this issue.\n\n'
                'A complete log of your simulator {} job will be available for 90 days [here]({}).{}\n\n'
                ).format(submitter, actions, not_actions, job_type, self.get_gh_action_run_url(), test_results_msg)

    def exec_core(self, submission, submitter):
        """ Validate simulator

        * Validate specifications
        * Validate image

        Args:
            submission (:obj:`SimulatorSubmission`): simulator submission
            submitter (:obj:`str`): GitHub id of the submitter

        Returns:
            :obj:`tuple`:

                * :obj:`dict`: specifications of a simulation tool
                * :obj:`list` of :obj:`TestCaseResults`: results of test cases
        """
        # validate specifications
        specifications = biosimulators_utils.simulator.io.read_simulator_specs(
            submission.specifications_url, submission.specifications_patch,
            validate=True)

        # check permissions
        self.validate_permissions(specifications['id'], specifications['name'], submitter)

        # indicate that specifications are valid
        self.add_comment_to_issue(self.issue_number, 'The specifications of your simulator is valid!')

        # validate image
        if submission.validate_image:
            test_results = self.validate_image(specifications)
            self.add_comment_to_issue(self.issue_number, 'The image for your simulator is valid!')
        else:
            test_results = None

        # return specifications
        return specifications, test_results

    def validate_permissions(self, simulator_id, simulator_name, submitter):
        """ Validate that the submitter has permissions to submit or update the simulator

        Args:
            simulator_id (:obj:`str`): simulator id
            simulator_name (:obj:`str`): simulator name
            submitter (:obj:`str`): GitHub id of the submitter
        """
        # check if team exists
        response = requests.get(
            'https://api.github.com/orgs/biosimulators/teams/' + simulator_id,
            auth=self.get_gh_auth())
        try:
            response.raise_for_status()
            has_team = True
        except requests.exceptions.HTTPError:
            if response.status_code == 404:
                has_team = False
            else:
                raise

        # create team if none exists and add submitter as a maintainer of the team
        if not has_team:
            # get parent team
            response = requests.get(
                'https://api.github.com/orgs/biosimulators/teams/simulator-developers',
                auth=self.get_gh_auth())
            response.raise_for_status()
            base_team_id = response.json()['id']

            # create team
            response = requests.post(
                'https://api.github.com/orgs/biosimulators/teams',
                auth=self.get_gh_auth(), json={
                    'name': simulator_id,
                    'description': 'Developers of ' + simulator_name,
                    'parent_team_id': base_team_id,
                    'maintainers': [submitter],
                })
            response.raise_for_status()

            # tell user a group was created
            msg = (
                'We created the GitHub group @biosimulators/{} to manage permissions to change the specifications of {} '
                'and added you (@{}) to this group. You can manage permissions to change the specifications of {} at '
                'https://github.com/orgs/biosimulators/teams/{}/members.'
            ).format(simulator_id, simulator_name, submitter, simulator_name, simulator_id)
            self.add_comment_to_issue(self.issue_number, msg)

        # check submitter has permissions to team
        else:
            response = requests.get(
                'https://api.github.com/orgs/biosimulators/teams/{}/memberships/{}'.format(simulator_id, submitter),
                auth=self.get_gh_auth())
            try:
                response.raise_for_status()
                has_permissions = True
            except requests.exceptions.HTTPError:
                if response.status_code == 404:
                    has_permissions = False
                else:
                    raise

            if not has_permissions:
                msg = (
                    'You (@{}) do not have permissions to update the specifications of {}. Only the members of @biosimulators/{} '
                    'can update the specifications of {}. Please contact the members of this group to request permissions to update {}.'
                ).format(submitter, simulator_name, simulator_id, simulator_name, simulator_name)
                self.add_error_comment_to_issue(self.issue_number, [Comment(text=msg, error=True)])

    def validate_image(self, specifications):
        """ Validate a Docker image for simulation tool

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool

        Returns:
            :obj:`list` of :obj:`TestCaseResults`: results of test cases
        """
        docker_client = biosimulators_utils.image.login_to_docker_registry(
            'docker.io', self.config.docker_hub_username, self.config.docker_hub_token)

        # validate that container (Docker image) exists
        image_url = specifications['image']['url']
        get_docker_image(docker_client, image_url, pull=True)

        # validate that Docker image can be converted to a Singularity image
        biosimulators_utils.image.convert_docker_image_to_singularity(
            image_url,
            singularity_filename=get_singularity_image_filename(image_url))

        # validate that image is consistent with the BioSimulators standards
        validator = SimulatorValidator(specifications, output_medium=OutputMedium.gh_issue)
        case_results = validator.run()
        write_test_results(case_results, '.biosimulators-test-suite-results.json',
                           gh_issue=int(self.issue_number), gh_action_run=int(self.get_gh_action_run_id()))
        summary, failure_details, warning_details, skipped_details = validator.summarize_results(
            case_results, output_medium=OutputMedium.gh_issue)

        # print summary to console
        print('')
        print('=============== SUMMARY ===============')
        print('')
        print(summary + '\n\n')
        if failure_details:
            color = Colors.failure.value
            print(termcolor.colored('=============== FAILURES ===============', color))
            print(termcolor.colored('', color))
            print(termcolor.colored('* ' + '\n\n* '.join(failure_details), color))
            print('')
        if warning_details:
            color = Colors.warning.value
            print(termcolor.colored('=============== WARNINGS ===============', color))
            print(termcolor.colored('', color))
            print(termcolor.colored('* ' + '\n\n* '.join(warning_details), color))
            print('')
        if skipped_details:
            color = Colors.skipped.value
            print(termcolor.colored('================ SKIPS =================', color))
            print(termcolor.colored('', color))
            print(termcolor.colored('* ' + '\n\n* '.join(skipped_details), color))
            print('')

        # push summary to comments on GitHub issue
        unable_to_post_results_msg = (
            'The summary of the tests of your Docker image could not be posted to this GitHub issue. '
            'The most likely reason is that the summary is too long to post to a comment on a GitHub issue. '
            'Please use the [console log]({}) of the associated GitHub Action to see the summary of the tests of your '
            'Docker image.'
        ).format(self.get_gh_action_run_url())

        msg = '## Summary of tests\n\n{}\n\n'.format(summary)
        self.add_comment_to_issue(self.issue_number, msg, alternative_comment=unable_to_post_results_msg)

        if failure_details:
            msg = '\n## Failures\n\n{}\n\n'.format('\n\n'.join(failure_details))
            self.add_comment_to_issue(self.issue_number, msg, alternative_comment=unable_to_post_results_msg)

        if warning_details:
            msg = '\n## Warnings\n\n{}\n\n'.format('\n\n'.join(warning_details))
            self.add_comment_to_issue(self.issue_number, msg, alternative_comment=unable_to_post_results_msg)

        if skipped_details:
            msg = '\n## Skips\n\n{}\n\n'.format('\n\n'.join(skipped_details))
            self.add_comment_to_issue(self.issue_number, msg, alternative_comment=unable_to_post_results_msg)

        invalid_cases = [case_result for case_result in case_results if case_result.type == TestCaseResultType.failed]
        if invalid_cases:
            gh_action_run_url = self.get_gh_action_run_url()
            error_msg = (
                'After correcting your simulator, please edit the first block of this issue to re-initiate this validation.\n\n'
                'The complete log of your validation/submission job, including further information about the failure, '
                'is available for 90 days [here]({}). The results of the validation of your image will also be '
                'available shortly as a JSON file. A link to this file will be available from the "Artifacts" '
                'section at the bottom of [this page]({}).'
            ).format(gh_action_run_url, gh_action_run_url)

            self.add_error_comment_to_issue(self.issue_number, [Comment(text=error_msg, error=True)])

        valid_cases = [case_result for case_result in case_results
                       if case_result.type == TestCaseResultType.passed and case_result.case.id.startswith('sedml.')]
        if not valid_cases:
            self.add_error_comment_to_issue(self.issue_number, [Comment(text=(
                'No test cases are applicable to your simulator. '
                'Please use this issue to share appropriate test COMBINE/OMEX files. '
                'The BioSimulators Team will add these files to this validation program and then re-review your simulator.'
            ))])

        return case_results

    def is_simulator_approved(self, specifications, existing_version_specifications):
        """ Determine whether a simulation tool has already been approved

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool
            existing_version_specifications (:obj:`list` of :obj:`dict`): specifications of other versions of simulation tool

        Returns:
            :obj:`bool`: :obj:`True`, if the simulation tool has already been approved
        """
        return (self.does_submission_have_approved_label()
                or self.is_other_version_of_simulator_validated(specifications, existing_version_specifications))

    def does_submission_have_approved_label(self):
        """ Determine whether an issue for submitting a simulator already has the approved label

        Returns:
            :obj:`bool`: :obj:`True`, if the issue for the submission already has the approved label
        """
        if IssueLabel.approved.value in self.get_labels_for_issue(self.issue_number):
            return True

    def is_other_version_of_simulator_validated(self, specifications, existing_version_specifications):
        """ Determine whether another version of the simulation tool has already been approved

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool
            existing_version_specifications (:obj:`list` of :obj:`dict`): specifications of other versions of simulation tool

        Returns:
            :obj:`bool`: :obj:`True`, if the simulation tool has already been approved
        """
        for version_spec in existing_version_specifications:
            if version_spec.get('biosimulators', {}).get('validated', False):
                return True
        return False

    def commit_simulator(self, submission, specifications, existing_version_specifications, test_results):
        """ Commit simulator to the BioSimulators registry

        Args:
            submission (:obj:`SimulatorSubmission`): simulator submission
            specifications (:obj:`dict`): specifications of a simulation tool
            existing_version_specifications (:obj:`list` of :obj:`dict`): specifications of other versions of simulation tool
            test_results (:obj:`list` of :obj:`TestCaseResults`): results of test cases
        """
        # commit image
        if submission.validate_image:
            # copy image to BioSimulators namespace of Docker registry (GitHub Container Registry)
            self.push_image(specifications, existing_version_specifications)

        # commit submission to BioSimulators database
        if 'biosimulators' not in specifications:
            specifications['biosimulators'] = {}

        if submission.validate_image:
            specifications['biosimulators']['validated'] = True
            specifications['biosimulators']['validationTests'] = TestResultsReport(
                results=test_results, gh_issue=int(self.issue_number), gh_action_run=int(self.get_gh_action_run_id()),
            ).to_dict(max_log_len=2**16)

        else:
            specifications['biosimulators']['validated'] = False
            specifications['biosimulators']['validationTests'] = None

        self.post_entry_to_biosimulators_api(specifications, existing_version_specifications)

        # commit image
        if submission.validate_image:
            # instruct runBioSimulations to generate a Singularity image for the Docker image
            self.trigger_conversion_of_docker_image_to_singularity(specifications)

    def push_image(self, specifications, existing_version_specifications):
        """ Push the image for a simulation tool to the GitHub Container Registry

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool
            existing_version_specifications (:obj:`list` of :obj:`dict`): specifications of other versions of simulation tool
        """
        # get image, pulling if necessary
        original_image_url = specifications['image']['url']
        docker_client = biosimulators_utils.image.login_to_docker_registry(
            'docker.io', self.config.docker_hub_username, self.config.docker_hub_token)
        image = get_docker_image(docker_client, original_image_url, pull=True)

        # push image to BioSimulators namespace of Docker registry
        biosimulators_utils.image.login_to_docker_registry(
            self.config.biosimulators_docker_registry_url,
            self.config.biosimulators_docker_registry_username,
            self.config.biosimulators_docker_registry_token)

        copy_image_url = self.config.biosimulators_docker_image_url_pattern \
            .format(specifications['id'], specifications['version']) \
            .lower()
        biosimulators_utils.image.tag_and_push_docker_image(docker_client, image, copy_image_url)
        specifications['image']['url'] = copy_image_url

        image = get_docker_image(docker_client, copy_image_url, pull=False)
        specifications['image']['digest'] = image.attrs['RepoDigests'][0].partition('@')[2]

        is_latest = self.is_submission_latest_version_of_simulator(specifications, existing_version_specifications)

        if is_latest:
            latest_copy_image_url = self.config.biosimulators_docker_image_url_pattern \
                .format(specifications['id'], 'latest') \
                .lower()
            biosimulators_utils.image.tag_and_push_docker_image(docker_client, image, latest_copy_image_url)

        # make image public -- must be done manually; cannot be done via API as of 2020-11-11

    def is_submission_latest_version_of_simulator(self, specifications, existing_version_specifications):
        """ Determine whether the submitted version the latest version of the simulator (greatest tag)

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool
            existing_version_specifications (:obj:`list` of :obj:`dict`): specifications of other versions of simulation tool

        Returns:
            :obj:`bool`: :obj:`True` if the submitted version if the latest version of the simulator
        """
        version_comparison_func = natsort_keygen()
        for existing_version_spec in existing_version_specifications:
            if (
                existing_version_spec['image']
                and existing_version_spec.get('biosimulators', {}).get('validated', False)
                and version_comparison_func(existing_version_spec['version']) > version_comparison_func(specifications['version'])
            ):
                return False
        return True

    def trigger_conversion_of_docker_image_to_singularity(self, specifications):
        """ Post the simulation to the BioSimulators database

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool
        """
        auth_headers = self.get_auth_headers_for_biosimulations_api(
            self.config.runbiosimulations_auth_endpoint, self.config.runbiosimulations_audience,
            self.config.runbiosimulations_api_client_id, self.config.runbiosimulations_api_client_secret)

        response = requests.post(self.config.runbiosimulations_api_endpoint + 'images/refresh',
                                 headers=auth_headers,
                                 json={
                                     'simulator': specifications['id'],
                                     'version': specifications['version'],
                                 })
        validate_biosimulations_api_response(response, 'A Singularity image could not be generated for the Docker image')

    def post_entry_to_biosimulators_api(self, specifications, existing_version_specifications):
        """ Post the simulation to the BioSimulators database

        Args:
            specifications (:obj:`dict`): specifications of a simulation tool
            existing_version_specifications (:obj:`list` of :obj:`dict`): specifications of other versions of simulation tool
        """
        auth_headers = self.get_auth_headers_for_biosimulations_api(
            self.config.biosimulators_auth_endpoint, self.config.biosimulators_audience,
            self.config.biosimulators_api_client_id, self.config.biosimulators_api_client_secret)

        existing_versions = [existing_version_spec['version'] for existing_version_spec in existing_version_specifications]
        update_simulator = specifications['version'] in existing_versions
        if update_simulator:
            endpoint = '{}simulators/{}/{}'.format(self.config.biosimulators_api_endpoint, specifications['id'], specifications['version'])
            requests_method = requests.put
            method = 'updated'
        else:
            endpoint = '{}simulators'.format(self.config.biosimulators_api_endpoint)
            requests_method = requests.post
            method = 'added'
        response = requests_method(endpoint, headers=auth_headers, json=specifications)
        validate_biosimulations_api_response(response, 'Simulation tool `{}` could not be {} to the BioSimulators registry.'.format(
            specifications['id'], method))

    def get_auth_headers_for_biosimulations_api(self, endpoint, audience, client_id, client_secret):
        """ Get authorization headers for using one of the BioSimulations REST APIs
        (BioSimulators, runBioSimulations, etc.).

        Args:
            endpoint (:obj:`str`): URL for getting an authentication token
            audience (:obj:`str`): API audience
            client_id (:obj:`str`): id for this client
            client_secret (:obj:`str`): secret for this client

        Returns:
            :obj:`dict`: authorization headers
        """
        response = requests.post(endpoint, json={
            'client_id': client_id,
            'client_secret': client_secret,
            'audience': audience,
            "grant_type": "client_credentials",
        })
        response.raise_for_status()
        response_data = response.json()
        return {'Authorization': response_data['token_type'] + ' ' + response_data['access_token']}

    @classmethod
    def add_comment_to_issue(cls, issue_number, comment, alternative_comment=None, max_len=65536):
        """ Post a comment to the GitHub issue

        Args:
            issue_number (:obj:`str`): issue number
            comment (:obj:`str`): comment
            alternative_comment (:obj:`str`, optional): optional alternative comment to try posting to the GitHub issue
            max_len (:obj:`int`, optional): maximum comment length accepted by GitHub
        """
        try:
            if len(comment) > max_len:
                comment = comment[0:max_len - 4] + ' ...'
            super(ValidateCommitSimulatorGitHubAction, cls).add_comment_to_issue(issue_number, comment)

        except (requests.exceptions.RequestException, requests.exceptions.ProxyError, requests.exceptions.SSLError):
            if alternative_comment:
                super(ValidateCommitSimulatorGitHubAction, cls).add_comment_to_issue(issue_number, alternative_comment)
            else:
                raise



""" Data model for results of test cases

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-01
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .._version import __version__
from ..warnings import TestCaseWarning  # noqa: F401
import enum
import traceback

__all__ = [
    'TestCaseResultType',
    'TestCaseResult',
    'TestResultsReport',
]


class TestCaseResultType(str, enum.Enum):
    """ Type of test case result """
    passed = 'passed'
    failed = 'failed'
    skipped = 'skipped'


class TestCaseResult(object):
    """ A result of executing a test case

    Attributes:
        case (:obj:`TestCase`): test case
        type (:obj:`obj:`TestCaseResultType`): type
        duration (:obj:`float`): execution duration in seconds
        exception (:obj:`Exception`): exception
        exception_traceback (:obj:`str`): traceback
        warnings (:obj:`list` of :obj:`TestCaseWarning`): warnings
        skip_reason (:obj:`Exception`): Exception which explains reason for skip
        log (:obj:`str`): log of execution
    """

    def __init__(self, case=None, type=None, duration=None, exception=None, exception_traceback=None, warnings=None, skip_reason=None, log=None):
        """
        Args:
            case (:obj:`TestCase`, optional): test case
            type (:obj:`obj:`TestCaseResultType`, optional): type
            duration (:obj:`float`, optional): execution duration in seconds
            exception (:obj:`Exception`, optional): exception
            exception_traceback (:obj:`str`, optional): traceback
            warnings (:obj:`list` of :obj:`TestCaseWarning`, optional): warnings
            skip_reason (:obj:`Exception`, optional): Exception which explains reason for skip
            log (:obj:`str`, optional): log of execution
        """
        self.case = case
        self.type = type
        self.duration = duration
        self.exception = exception
        self.exception_traceback = exception_traceback
        self.warnings = warnings or []
        self.skip_reason = skip_reason
        self.log = log

    def to_dict(self, max_log_len=None, debug=True):
        """ Generate a dictionary representation e.g., for export to JSON

        Args:
            max_log_len (:obj:`int`, optional): maximum log length
            debug (:obj:`bool`, optional): whether to display traceback information about each error
                with additional information for debugging

        Returns:
            :obj:`dict`: dictionary representation
        """
        log = self.log
        if max_log_len is not None and log and len(log) > max_log_len:
            log = log[0:max_log_len - 1] + ' ...'

        return {
            'case': {
                'id': self.case.id,
                'description': self.case.description,
            },
            'resultType': self.type.value,
            'duration': self.duration,
            'exception': {
                'category': self.exception.__class__.__name__,
                'message': str(self.exception),
                'traceback': traceback.format_tb(self.exception_traceback) if self.exception_traceback else None,
            } if self.exception else None,
            'warnings': [{'category': warning.category.__name__, 'message': str(warning.message)}
                         for warning in self.warnings],
            'skipReason': {
                'category': self.skip_reason.__class__.__name__,
                'message': str(self.skip_reason),
            } if self.skip_reason else None,
            'log': log,
        }


class TestResultsReport(object):
    """ A report of the results of executing the test suite with a simulation tool

    Attributes:
        test_suite_version (:obj:`str`): version of the test suite which was executed
        results (:obj:`list` of :obj:`TestCaseResult`): results of the test cases of the test suite
        gh_issue (:obj:`int`): GitHub issue for which the test suite was executed
        gh_action_run (:obj:`int`): GitHub action run in which the test suite was executed
    """

    def __init__(self, test_suite_version=__version__, results=None, gh_issue=None, gh_action_run=None):
        """
        Args:
            test_suite_version (:obj:`str`, optional): version of the test suite which was executed
            results (:obj:`list` of :obj:`TestCaseResult`, optional): results of the test cases of the test suite
            gh_issue (:obj:`int`, optional): GitHub issue for which the test suite was executed
            gh_action_run (:obj:`int`, optional): GitHub action run in which the test suite was executed
        """
        self.test_suite_version = test_suite_version
        self.results = results or []
        self.gh_issue = gh_issue
        self.gh_action_run = gh_action_run

    def to_dict(self, max_log_len=None):
        """ Generate a dictionary representation e.g., for export to JSON

        Args:
            max_log_len (:obj:`int`, optional): maximum log length

        Returns:
            :obj:`dict`: dictionary representation
        """
        return {
            'testSuiteVersion': self.test_suite_version,
            'results': [result.to_dict(max_log_len=max_log_len) for result in self.results],
            'ghIssue': self.gh_issue,
            'ghActionRun': self.gh_action_run,
        }



""" Data model for test cases for validating simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .config import Config
from .exceptions import SkippedTestCaseException  # noqa: F401
from biosimulators_utils.image import get_docker_image
import abc
import docker
import enum

__all__ = [
    'OutputMedium',
    'TestCase', 'SedTaskRequirements', 'ExpectedSedReport', 'ExpectedSedDataSet', 'ExpectedSedPlot',
    'AlertType',
]


class OutputMedium(str, enum.Enum):
    """ Output medium """
    console = 'console'
    gh_issue = 'gh_issue'


class TestCase(abc.ABC):
    """ A test case for validating a simulator

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        description (:obj:`str`): description
        output_medium (:obj:`OutputMedium`): medium the description should be formatted for
    """

    def __init__(self, id=None, name=None, description=None, output_medium=OutputMedium.console):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            description (:obj:`str`): description
            output_medium (:obj:`OutputMedium`, optional): medium the description should be formatted
        """
        self.id = id
        self.name = name
        self.description = description
        self.output_medium = output_medium

    @abc.abstractmethod
    def eval(self, specifications, working_dirname, synthetic_archives_dir=None, dry_run=False, cli=None):
        """ Evaluate a simulator's performance on a test case

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            working_dirname (:obj:`str`): directory for temporary files for evaluating test case
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives
                generated by the test cases
            dry_run (:obj:`bool`, optional): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image

        Raises:
            :obj:`SkippedTestCaseException`: if the test case is not applicable to the simulator
            :obj:`Exception`: if the simulator did not pass the test case
        """
        pass  # pragma: no cover

    def get_simulator_docker_image(self, specifications, pull=None):
        """ Get the Docker image for a simulator, pulling if necessary

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate

        Returns:
            :obj:`docker.models.images.Image`: Docker image
        """
        docker_client = docker.from_env()
        image_url = specifications['image']['url']
        if pull is None:
            pull = Config().pull_docker_image
        return get_docker_image(docker_client, image_url, pull=pull)


class SedTaskRequirements(object):
    """ Required model format for simulation algorithm for each task in a SED document

    Attributes:
        model_format (:obj:`str`): EDAM id for the format of the model involved in the task
        model_format_features (:obj:`set` of :obj:`str`): model format features required to execute the task
                such as SBML packages
        simulation_algorithm (:obj:`str`): KiSAO id for the simulation algorithm involved in the task
    """

    def __init__(self, model_format=None, model_format_features=None, simulation_algorithm=None):
        """
        Args:
            model_format (:obj:`str`, optional): EDAM id for the format of the model involved in the task
            model_format_features (:obj:`set` of :obj:`str`, optional): model format features required to execute the task
                such as SBML packages
            simulation_algorithm (:obj:`str`, optional): KiSAO id for the simulation algorithm involved in the task
        """
        self.model_format = model_format
        self.model_format_features = model_format_features or set()
        self.simulation_algorithm = simulation_algorithm


class ExpectedSedReport(object):
    """ An expected SED report

    Attributes
        id (:obj:`str`): id
        data_sets (:obj:`list` of :obj:`ExpectedSedDataSet`): labels of expected data sets
        points (:obj:`tuple` of :obj:`int`): number of expected points of
        values (:obj:`dict` of :obj:`str` to :obj:`dict` of :obj:`list`): expected values of data sets or elements of data sets
    """

    def __init__(self, id=None, data_sets=None, points=None, values=None):
        """
        Args:
            id (:obj:`str`, optional): id
            data_sets (:obj:`set` of :obj:`ExpectedSedDataSet`, optional): labels of expected data sets
            points (:obj:`tuple` of :obj:`int`, optional): number of expected points of
            values (:obj:`dict` of :obj:`str` to :obj:`dict` of :obj:`list`, optional): expected values of data sets or elements of data sets
        """
        self.id = id
        self.data_sets = data_sets or set()
        self.points = points
        self.values = values


class ExpectedSedDataSet(object):
    """ An expected SED report

    Attributes
        id (:obj:`str`): id
        label (:obj:`str`): label
    """

    def __init__(self, id=None, label=None):
        """
        Args:
            id (:obj:`str`): id
            label (:obj:`str`): label
        """
        self.id = id
        self.label = label


class ExpectedSedPlot(object):
    """ An expected SED report

    Attributes
        id (:obj:`str`): id
    """

    def __init__(self, id=None):
        """
        Args:
            id (:obj:`str`, optional): id
        """
        self.id = id


class AlertType(str, enum.Enum):
    """ Type of alert upon the failure of a test case """
    exception = 'exception'
    warning = 'warning'
